use std::{collections::HashMap, fmt::Display, fs, sync::Arc};

use crate::Context;
use anyhow::Error;
use poise::{serenity_prelude::{Cache, CacheHttp, ChannelId, ChannelType, GetMessages, GuildChannel, Http, Message}, CreateReply};
use serde::Serialize;
use tokio::time::Instant;
use tracing::{debug, error, info, trace};

#[derive(Serialize)]
struct Server {
    channels: Vec<Channel>,
    orphanage: Vec<GuildChannel>,
    needs_clean: bool,
}

#[derive(Serialize)]
struct Channel {
    this: GuildChannel,
    children: Vec<Channel>,
    messages: Vec<Message>,
}

impl Channel {
    fn new(this: GuildChannel) -> Self {
        Self {
            this,
            // Empty vecs don't allocate until a push
            children: Vec::new(),
            messages: Vec::new(),
        }
    }
}

impl Display for Server {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {

        fn print(f: &mut std::fmt::Formatter<'_>, tab: usize, channel: &Vec<Channel>) {
            for i in channel {
                for _ in 0..tab { let _ = write!(f, "\t"); }
                let _ = writeln!(f, "{} {}", prefix(i.this.kind),i.this.name);
                print(f, tab+1, &i.children);
            }
        }

        fn prefix(kind: ChannelType) -> &'static str {
            match kind {
                poise::serenity_prelude::ChannelType::Text => "💬",
                poise::serenity_prelude::ChannelType::Private => "🕵️",
                poise::serenity_prelude::ChannelType::Voice => "🔊",
                poise::serenity_prelude::ChannelType::GroupDm => "👨‍👦‍👦",
                poise::serenity_prelude::ChannelType::Category => "📁",
                poise::serenity_prelude::ChannelType::News => "📢",
                poise::serenity_prelude::ChannelType::NewsThread => "📢🧵",
                poise::serenity_prelude::ChannelType::PublicThread => "🧵",
                poise::serenity_prelude::ChannelType::PrivateThread => "🕵️🧵",
                poise::serenity_prelude::ChannelType::Stage => "🎙️",
                poise::serenity_prelude::ChannelType::Directory => "📒",
                poise::serenity_prelude::ChannelType::Forum => "🗣️",
                _ => "?",
            }
        }

        print(f, 0, &self.channels); 


        if self.needs_clean {
            let _ = writeln!(f, "Orphans: (please clean() before displaying...)");
            for i in &self.orphanage {
                let _ = write!(f, "{} {},", prefix(i.kind),i.name);
            }
        }

        write!(f, "")
    }
}

impl Server {

    fn new() -> Self {
        Self {
            channels: Vec::new(),
            orphanage: Vec::new(), 
            needs_clean: false
        }
    }
    
    fn search_by_id<'a>(target: &'a mut Vec<Channel>, find: &ChannelId) -> Option<&'a mut Channel> {
        for child in target {
            if child.this.id == *find {
                return Some(child);
            }
            if let Some(x) = Self::search_by_id(&mut child.children, find) {
                return Some(x);
            }
       }
        None
   }

    fn add(&mut self, insert: GuildChannel) {
        // make sure the new item wants a parent
        if let Some(parent_id) = &insert.parent_id {
            // find the parent (needs to go thru all nodes)

            match Self::search_by_id(&mut self.channels, parent_id) {
                Some(parent_node) => {
                    parent_node.children.push(Channel::new(insert));
                },
                None => {
                    // couldn't find parent, store somewhere else until it's parent is added...
                    self.orphanage.push(insert);
                    self.needs_clean = true;
                },
            }
        } else {
            self.channels.push(Channel::new(insert));
        }
    }

    /// Cleans out the orphan channels, finding them parents. You'll want to use this before displaying anything.
    fn clean(&mut self) {
        if !self.needs_clean {return;}

        // Look thru the orphanage and try to find parents
        for orphan in &self.orphanage {
            if let Some(parent_id) = orphan.parent_id {
                if let Some(found) = Self::search_by_id(&mut self.channels, &parent_id) {
                    found.children.push(Channel::new(orphan.clone()));
                } else {
                    panic!("⚠️ Couldn't find parent for orphan!");
                }
            } else {
                panic!("⚠️ All orphans should want a parent node!");
            }
        }
        self.orphanage.clear();
        self.needs_clean = false;
    }

    /// Scrapes messages for all the channels in `self`.
    async fn scrape_all(&mut self) {
        // TODO add status from this command, just because it's cool
        let cache: (&Arc<Cache>, &Http) = (&Arc::new(Cache::new()), &Http::new(&crate::ENV.token));
        walk_channels(&mut self.channels, cache).await;

        /// Recursive walk thru the channels
        async fn walk_channels(all: &mut Vec<Channel>, cache: impl CacheHttp + Clone) {
            // Qty of messages to take at a time, max=100
            let batch_size = 100;

            let settings = GetMessages::default().limit(batch_size);
            for channel in all {
                // Clone *should* be cheap - it's Arc under the hood
                // get the messages
                get_messages(channel, cache.clone(), settings).await;
            }           

            /// Get all messages for 1 channel and children
            async fn get_messages(channel: &mut Channel, cache: impl CacheHttp + Clone, settings: GetMessages) {
                // Loop thru all the messages in the channel in batches.
                // Adding each batch to the current channel's messages the whole time.
                let mut last_id = channel.this.last_message_id;
                while let Some(last) = last_id {
                    match channel.this.messages(cache.clone(), settings.before(last)).await {
                        Ok(mut ok) => {
                            if ok.is_empty() {
                                debug!("Reached the beginning of {}", channel.this.name);
                                // Stop the loop if there are no more messages.
                                last_id = None;  
                            } else {
                                trace!("Adding {} messages to \"{}\"", ok.len(), channel.this.name);
                                channel.messages.append(&mut ok);
                                // extract message id
                                if let Some(l) = channel.messages.last() {
                                    last_id = Some(l.id);
                                }
                            }                        
                        },
                        Err(e) => {
                            error!("While reading messages in \"{}\" before `{}` - {e}", channel.this.name, last);
                            // Stop reading this channel on an error.
                            last_id = None;
                        },
                    }
                }
                // Then recurse into children channels
                Box::pin(walk_channels(&mut channel.children, cache.clone())).await;
            }
        }
    }

    /// Walk thru all the channels and count the saved messages. Will only give relevant data if
    /// done after `scrape_all()`.
    fn message_count(&self) -> usize {
        fn walk(this: &Vec<Channel>) -> usize {
            let mut total = 0;
            for channel in this {
                total += channel.messages.len();
                total += walk(&channel.children);
            };
            total
        }
        walk(&self.channels)
    }

}

#[poise::command(slash_command, rename = "scrape_all", guild_only, owners_only)]
pub async fn scrape_all(ctx: Context<'_>, pretty_print: bool) -> Result<(), Error> {
    let guild = ctx.guild_id().unwrap().to_partial_guild(ctx.serenity_context()).await.unwrap();

    let invoker = ctx.author().name.clone();
    if let Some(nickname) = ctx.author().nick_in(ctx.http(), guild.id).await {
        info!("{invoker} ({nickname}) started a scrape of {}", guild.name);
    } else {
        info!("{invoker} started a scrape of {}", guild.name);
    }

    if let Ok(map) = guild.channels(ctx.http()).await {
        let mut server = index(map).await;
        match ctx.reply("Starting scrape...").await {
            Ok(ok) => {
                let start = Instant::now();
                server.scrape_all().await;
                let end = start.elapsed().as_millis();
                let msg_count = server.message_count();

                let print = if pretty_print {
                    serde_json::to_string_pretty(&server)
                } else {
                    serde_json::to_string(&server)
                };

                match print {
                    Ok(ok) => {
                        if let Err(e) = fs::write("server.json", ok) {
                            error!("Problem writing server to disk: {e}");
                        }
                    },
                    Err(err) => {
                        error!("Trying to serialize server: {err}");
                    },
                }

                // Done. Print stats.
                let _ = ok.edit(ctx, CreateReply::default().content(
                    &format!("Done. Stats: \n```toml\nMessages saved: {msg_count}\nElapsed time: {end}ms\n```")
                )).await;
                debug!("Scraped server in {}ms", end);
            },
            Err(e) => error!("{e} - While trying to reply to scrape command"),
        }
   }
    Ok(())
}

/// Get server's topology (and runs clean)
async fn index(map: HashMap<ChannelId, GuildChannel>) -> Server {
    let mut server = Server::new();
    // iterate thru all channels
    map.into_iter().for_each(|(_id, current)| {
        // println!("{} {} {:?}", current.name, current.id, current.parent_id);
        server.add(current);
        // TODO Take note of position
        // TODO Take node of vc user limit
    });
    server.clean();
    server
}

// NOTE!!! Make sure these names in quotes are lowercase!
#[poise::command(slash_command, rename = "index", guild_only)]
pub async fn index_cmd(ctx: Context<'_>) -> Result<(), Error> {
    let guild = ctx.guild_id().unwrap().to_partial_guild(ctx.serenity_context()).await.unwrap();
    let invoker = ctx.author().name.clone();
    if let Some(nickname) = ctx.author().nick_in(ctx.http(), guild.id).await {
        info!("{invoker} ({nickname}) is indexing {}", guild.name);
    } else {
        info!("{invoker} is indexing {}", guild.name);
    }

    match guild.channels(ctx.http()).await {
        Ok(ok) => {
            let server = index(ok).await;
            let _ = ctx.reply(server.to_string()).await;
        },
        Err(_) => todo!(),
    }
    Ok(())
}
